#include "GLFW/glfw3.h"
#include "cglm/cam.h"
#include "cglm/mat4.h"
#include "ui.h"
#include "gpu.h"
#include "draw.h"
#include "hex.h"

#include "arpa/inet.h"
#include "vk_mem_alloc.h"
#include "vulkan/vk_enum_string_helper.h"
#include "vulkan/vulkan_core.h"
#include "pthread.h"
#include <math.h>

typedef struct ClientContextStruct {
  GLFWwindow* window;
  RenderContext render;
  UIContext ui;

  uint32_t clicked_container;
  uint32_t clicked_element;

  vec2    cursor;

  vec3    position;

  vec2    rotation;
  int32_t key_spin[2];
  vec2    cur_spin;

  double  distance;
  int32_t zoom;

  bool    camera_mode;
  float   key_spin_speed;
  float   cur_spin_speed;
  float   zoom_speed;
} ClientContext;

void* network_thread(void* data) {
  ClientContext* context = (ClientContext*)data;
  (void)context;

  return NULL;
}

vec3 zero = {0, 0, 0};
vec3 up = {0, 1, 0};

VkResult main_thread(ClientContext* context) {
  VkResult result;

  GraphicsPipeline graphics;
  ComputePipeline compute;
  VK_RESULT(create_hex_pipeline(context->render.device, context->render.render_pass, &graphics, &compute));

  GPUString strings[] = {
    {
      .pos = {0, 32},
      .size = 32,
      .color = {1.0, 1.0, 1.0, 1.0},
      .offset = 0,
      .length = 0,
      .font = 0,
    },
    {
      .pos = {110, 138},
      .size = 32,
      .color = {1.0, 1.0, 1.0, 1.0},
      .offset = 20,
      .length = 1,
      .font = 0,
    },
    {
      .pos = {210, 138},
      .size = 32,
      .color = {1.0, 1.0, 1.0, 1.0},
      .offset = 21,
      .length = 1,
      .font = 0,
    },
  };

  GPUDrawable drawables[] = {
    {
      .pos = {100, 100},
      .size = {50, 50},
      .color = {1, 0, 0, 1},
      .type = 0,
      .id = 1,
    },
    {
      .pos = {200, 100},
      .size = {50, 50},
      .color = {0, 1, 0, 1},
      .type = 0,
      .id = 2,
    },
  };

  LayerInput layer = {
    .strings = strings,
    .num_strings = sizeof(strings)/sizeof(GPUString),
    .max_codes = 24,

    .drawables = drawables,
    .num_drawables = sizeof(drawables)/sizeof(GPUDrawable),
  };

  ContainerInput container = {
    .id = 1,
    .size = {WINDOW_MIN_WIDTH, WINDOW_MIN_HEIGHT},
    .layer_count = 1,
    .layers = &layer,
  };

  create_container(&container, &context->render, &context->ui);
  map_string("1234", context->ui.containers[0].layers[0].codes_buffer, 20, 0, &context->ui);
  VK_RESULT(add_transfers(
        &context->ui.containers[0].layers[0].codes_buffer[20],
        context->ui.containers[0].layers[0].codes,
        20*sizeof(uint32_t),
        4*sizeof(uint32_t),
        &context->render));

  VkBuffer hex_context;
  VmaAllocation hex_context_memory;
  GPUHexContext hex;
  VK_RESULT(create_storage_buffer(context->render.allocator, 0, sizeof(GPUHexContext), &hex_context, &hex_context_memory));
  VkDeviceAddress hex_context_address = buffer_address(context->render.device, hex_context);
  glm_perspective(30.0, (float)context->render.swapchain_extent.width/(float)context->render.swapchain_extent.height, 0.01, 99.9, hex.proj);
  context->position[0] = 0;
  context->position[1] = 0;
  context->position[2] = 0;

  context->rotation[0] = 0;
  context->rotation[1] = 0;
  context->cur_spin[0] = 0;
  context->cur_spin[1] = 0;

  context->key_spin[0] = 0;
  context->key_spin[1] = 0;
  context->cur_spin[0] = 0;
  context->cur_spin[1] = 0;
  context->distance = 1;
  context->zoom = 0;
  context->camera_mode = false;
  context->key_spin_speed = 1.0;
  context->cur_spin_speed = 1.0;
  context->zoom_speed = 1.0;

  //
  uint32_t* mapped_codes = context->ui.containers[0].layers[0].codes_buffer;
  GPUString* mapped_string = context->ui.containers[0].layers[0].strings_buffer;
  char str[21];
  //

  double last_frame_time = glfwGetTime();
  while(glfwWindowShouldClose(context->window) == 0) {
    double frame_time = glfwGetTime();
    double delta_time = (frame_time - last_frame_time);

    context->clicked_element   = 0x00000000;
    context->clicked_container = 0x00000000;
    context->zoom = 0;
    context->cur_spin[0] = 0;
    context->cur_spin[1] = 0;

    glfwPollEvents();

    if(context->key_spin[0] != 0 || context->key_spin[1] != 0 ||
       context->zoom != 0 ||
       context->cur_spin[0] != 0 || context->cur_spin[1] != 0) {

      context->rotation[0] += (float)context->key_spin[0]*delta_time*context->key_spin_speed; 
      context->rotation[0] += (float)context->cur_spin[0]*delta_time*context->cur_spin_speed;
      if(context->rotation[0] > 2*M_PI) {
        context->rotation[0] -= 2*M_PI;
      } else if(context->rotation[0] < 0) {
        context->rotation[0] += 2*M_PI;
      }

      context->rotation[1] += (float)context->key_spin[1]*delta_time*context->key_spin_speed;
      context->rotation[1] += (float)context->cur_spin[1]*delta_time*context->cur_spin_speed;
      if(context->rotation[1] > M_PI/2) {
        context->rotation[1] = M_PI/2;
      } else if(context->rotation[1] < 0) {
        context->rotation[1] = 0;
      }

      context->distance += context->zoom*delta_time*context->zoom_speed;
      if(context->distance < 1) {
        context->distance = 1;
      }

      vec3 camera;
      camera[0] = context->position[0] + context->distance*cos(context->rotation[1])*cos(context->rotation[0]);
      camera[1] = context->position[1] + context->distance*sin(context->rotation[1]);
      camera[2] = context->position[2] + context->distance*cos(context->rotation[1])*sin(context->rotation[0]);
      glm_lookat(camera, context->position, up, hex.view);

      add_transfer(&hex, hex_context, 0, sizeof(GPUHexContext), (context->render.current_frame + 1) % MAX_FRAMES_IN_FLIGHT, &context->render);
    }

    if(context->clicked_container != 0) {
      snprintf(str, 21, "Clicked: %d.%d", context->clicked_container, context->clicked_element);
      map_string(str, mapped_codes, 0, 0, &context->ui);
      VK_RESULT(add_transfers(
            context->ui.containers[0].layers[0].codes_buffer,
            context->ui.containers[0].layers[0].codes,
            0,
            strlen(str)*sizeof(uint32_t),
            &context->render));
      mapped_string->length = strlen(str);
      VK_RESULT(add_transfers(
            &context->ui.containers[0].layers[0].strings_buffer[0].length,
            context->ui.containers[0].layers[0].strings,
            offsetof(GPUString, length),
            sizeof(uint32_t),
            &context->render));

    }
    //

    VkResult result = draw_frame(&context->render, &context->ui, &graphics, hex_context_address, frame_time);
    if(result != VK_SUCCESS) {
      fprintf(stderr, "draw_frame error: %s\n", string_VkResult(result));
      glfwDestroyWindow(context->window);
    }

    last_frame_time = frame_time;
  }

  return 0;
}

void key_callback(GLFWwindow* window, int key, int scancode, int action, int mods) {
  ClientContext* context = (ClientContext*)glfwGetWindowUserPointer(window);
  (void)scancode;
  (void)mods;
  switch(key) {
  case GLFW_KEY_A:
    if(action == GLFW_PRESS) {
      context->key_spin[0] -= 1;
    } else if(action == GLFW_RELEASE) {
      context->key_spin[0] += 1;
    }
    break;
  case GLFW_KEY_D:
    if(action == GLFW_PRESS) {
      context->key_spin[0] += 1;
    } else if(action == GLFW_RELEASE) {
      context->key_spin[0] -= 1;
    }
    break;

  case GLFW_KEY_W:
    if(action == GLFW_PRESS) {
      context->key_spin[1] += 1;
    } else if(action == GLFW_RELEASE) {
      context->key_spin[1] -= 1;
    }
    break;
  case GLFW_KEY_S:
    if(action == GLFW_PRESS) {
      context->key_spin[1] -= 1;
    } else if(action == GLFW_RELEASE) {
      context->key_spin[1] += 1;
    }
    break;
  }
}

bool contains(double* point, GPUContainer* container, GPUDrawable* rect) {
  vec2 pos = {
    container->offset[0] + rect->pos[0],
    container->offset[1] + rect->pos[1],
  };
  return (point[0] >= pos[0] && point[0] <= pos[0] + rect->size[0]) &&
         (point[1] >= pos[1] && point[1] <= pos[1] + rect->size[1]) &&
         (point[0] >= container->offset[0] && point[0] <= container->offset[0] + container->size[0]) &&
         (point[1] >= container->offset[1] && point[1] <= container->offset[1] + container->size[1]);
}

void mouse_button_callback(GLFWwindow* window, int button, int action, int mods) {
  ClientContext* context = (ClientContext*)glfwGetWindowUserPointer(window);

  (void)mods;
  (void)context;
  double cursor[2];
  glfwGetCursorPos(window, &cursor[0], &cursor[1]);
  switch(button) {
  // Handle camera hover
  case GLFW_MOUSE_BUTTON_RIGHT:
    if(action == GLFW_PRESS) {
      glfwSetInputMode(window, GLFW_CURSOR, GLFW_CURSOR_DISABLED);
      context->camera_mode = true;
    } else {
      glfwSetInputMode(window, GLFW_CURSOR, GLFW_CURSOR_NORMAL);
      context->camera_mode = false;
    }
    break;
  case GLFW_MOUSE_BUTTON_LEFT:
    if(action == GLFW_PRESS) {
      for(int32_t c = context->ui.max_containers - 1; c >= 0; c--) {
        if(context->ui.containers[c].id == 0x00000000) {
          continue;
        }

        for(uint32_t l = 0; l < context->ui.containers[c].layer_count; l++) {
          for(uint32_t d = 0; d < context->ui.containers[c].layers[l].data.num_drawables; d++) {
            if(context->ui.containers[c].layers[l].drawables_buffer[d].id == 0x00000000) {
              continue;
            }
            if(contains(cursor, &context->ui.containers[c].data, &context->ui.containers[c].layers[l].drawables_buffer[d])) {
              context->clicked_container = context->ui.containers[c].id;
              context->clicked_element = context->ui.containers[c].layers[l].drawables_buffer[d].id;
            }
          }
        }
      }
      // NOTE: this logic is the same as it would be for right-click(just with different outcomes), so dont repeat yourself
      // 1. Search through the UI context for the first element that the contains the mouse point
      // 2. If no UI element intersection, cast a ray through the world scene to find the "clicked entity" 
      // 3. Based on what was clicked, start the mouse click animation at the target area
    }
    break;
  }
}

void scroll_callback(GLFWwindow* window, double xoffset, double yoffset) {
  ClientContext* context = (ClientContext*)glfwGetWindowUserPointer(window);
  (void)xoffset;
  context->zoom = -yoffset;  
}

void cursor_pos_callback(GLFWwindow* window, double xpos, double ypos) {
  ClientContext* context = (ClientContext*)glfwGetWindowUserPointer(window);
  if(context->camera_mode == true) {
    context->cur_spin[0] = (xpos - context->cursor[0])/context->render.swapchain_extent.width*-100;
    context->cur_spin[1] = (ypos - context->cursor[1])/context->render.swapchain_extent.height*100;

    context->cursor[0] = xpos;
    context->cursor[1] = ypos;
  }
}

int main() {
  ClientContext context = {};
  context.window = init_window();
  if(context.window == NULL) {
    return 1;
  }

  glfwSetWindowUserPointer(context.window, &context);
  glfwSetKeyCallback(context.window, key_callback);
  glfwSetMouseButtonCallback(context.window, mouse_button_callback);
  glfwSetScrollCallback(context.window, scroll_callback);
  glfwSetCursorPosCallback(context.window, cursor_pos_callback);

  int error;
  VkResult result;
  VK_RESULT(init_vulkan(context.window, &context.render));
  
  // TODO: make # of fonts/textures/containers scaling, recreate GPU buffers as necessary
  VK_RESULT(create_ui_context(10, 10, 10, &context.render, &context.ui));

  pthread_t network_thread_handle;

  error = pthread_create(&network_thread_handle, NULL, &network_thread, &context);
  if(error != 0) {
    return error;
  }

  error = main_thread(&context);
  if(error != 0) {
    return error;
  }

  error = pthread_join(network_thread_handle, NULL);
  if(error != 0) {
    return error;
  }

  return 0;
}
